movi	R6, 10
movi	R6, 10
movi	R6, 10
movi	R6, 10
movi	R6, 10
movi	R6, 10
movi	R5, 5
mov	R5, R5
mov	R5, R5
mov	R5, R5
mov	R5, R5
mov	R5, R5
addi	R5, R5, 1
bne     R5, R6, -6 
add	R1, R2, R3
add	R2, R3, R4
